import os

from src.unmeshed.sdk.common.process_data import ProcessData
from src.unmeshed.sdk.common.process_request_data import ProcessRequestData
from src.unmeshed.sdk.common.worker_profile import WorkerProfile
from src.unmeshed.sdk.configs.client_config import ClientConfig
from src.unmeshed.sdk.decorators.worker_function import worker_function
from src.unmeshed.sdk.unmeshed_client import UnmeshedClient
from src.unmeshed.sdk.apis.workers.worker import Worker
from src.unmeshed.sdk.common.api_call_type import ApiCallType
from src.unmeshed.sdk.common.process_search_request import ProcessSearchRequest
from src.unmeshed.sdk.common.step_data import StepData


class WorkerResponse:
    def __init__(self, success=False, message="", data=None, error_code=None, exception=None):
        """
        Initialize a WorkerResponse instance.

        :param success: Boolean indicating if the operation was successful
        :param message: Descriptive message about the response
        :param data: Payload/data returned by the worker (if any)
        :param error_code: Custom error code for classification
        :param exception: Exception object (if an exception occurred)
        """
        self.success = success
        self.message = message
        self.data = data
        self.error_code = error_code
        self.exception = exception

    def to_dict(self):
        """Convert the response to a dictionary format"""
        return {
            'success': self.success,
            'message': self.message,
            'data': self.data,
            'error_code': self.error_code,
            'exception_type': type(self.exception).__name__ if self.exception else None,
            'exception_message': str(self.exception) if self.exception else None
        }

    @classmethod
    def from_exception(cls, exception, error_code=None):
        """Create a WorkerResponse from an exception"""
        return cls(
            success=False,
            message=str(exception),
            data=None,
            error_code=error_code,
            exception=exception
        )

    def __repr__(self):
        return (f"WorkerResponse(success={self.success}, "
                f"message={self.message!r}, "
                f"data={self.data!r}, "
                f"error_code={self.error_code!r}, "
                f"exception={self.exception!r})")

    def __str__(self):
        if self.success:
            return f"Success: {self.message}"
        else:
            error_info = []
            if self.error_code:
                error_info.append(f"Code: {self.error_code}")
            if self.message:
                error_info.append(f"Message: {self.message}")
            return "Error - " + ", ".join(error_info)


@worker_function(name = "hello_world_python")
def task_hello_world(input_dict: dict) -> dict:
    print(f"Received input: {input_dict}")
    output_dict = {
        "message": "Hello, world!",
        "input_received": input_dict
    }
    return output_dict

@worker_function(name = "list_test_2")
def list_test(inp: dict) -> list:
    # Define a complex list with nested arrays and objects
    lst = [
        "23232",
        {
            "id": "1",
            "name": "Item 1",
            "tags": ["tag1", "tag2"],
            "details": {
                "description": "This is a description for item 1",
                "attributes": [100, 200, 300]
            }
        },
        {
            "id": "2",
            "name": "Item 2",
            "tags": ["tag3", "tag4"],
            "details": {
                "description": "This is a description for item 2",
                "attributes": [400, 500, 600]
            }
        },
        {
            "id": "3",
            "name": "Item 3",
            "tags": ["tag5", "tag6"],
            "details": {
                "description": "This is a description for item 3",
                "attributes": [700, 800, 900]
            }
        }
    ]
    return lst


def task_second_worker(response: WorkerResponse) -> WorkerResponse:
    print(f"Processing response: {response.to_dict()}")
    return WorkerResponse(
        success=True,
        message="Second worker processed",
        data={
            "original_response": response.to_dict(),
            "worker_note": "Processed by secondary worker"
        }
    )

class TestWorkerCallable:
    def __init__(self):
        pass
    @worker_function(name = "class_worker")
    def classWorker(self, input : dict) -> dict:
        print("Input received is " + str(input))
        return {
            "a": "bcd"
        }

class CustomError(RuntimeError):
    def __init__(self, message, error_code, error_data):
        super().__init__(message)
        self.error_code = error_code
        self.error_data = error_data

def exception_step(input_dict: dict) -> dict:
    """Raises a custom exception with sample error details"""
    raise CustomError(
        "Intentional exception from exception_step",
        error_code="CUSTOM_ERROR_123",
        error_data={"step": 2, "status": "failed"}
    )

def main():
    client_config = ClientConfig()
    client_config.set_client_id(os.getenv("CLIENT_ID"))
    client_config.set_auth_token(os.getenv("AUTH_TOKEN"))
    client_config.set_port(8080)
    client_config.set_work_request_batch_size(50)
    client_config.set_base_url("http://localhost")
    client_config.set_initial_delay_millis(50)
    client_config.set_step_timeout_millis(3600000)
    client_config.set_max_threads_count(10)

    client = UnmeshedClient(client_config)
    current_directory_full_path = os.getcwd()

    if os.getenv("get_process_data"):
        process_data: ProcessData = client.get_process_data(8100103, True)
        print(str(process_data))

    if os.getenv("get_step_data"):
        step_data: StepData = client.get_step_data(8100123)
        print(str(step_data))

    if os.getenv("bulk_terminate"):
      bulk_terminate_response = client.bulk_terminate([8150036, 8150039], "Testing from python SDK")
      print(str(bulk_terminate_response))

    if os.getenv("bulk_resume"):
      bulk_resume_response = client.bulk_resume([8150036, 8150039])
      print(str(bulk_resume_response))

    if os.getenv("bulk_review"):
        bulk_reviewed_response = client.bulk_reviewed([8150036, 8150039], "Reviewed test")
        print(str(bulk_reviewed_response))

    if os.getenv("run_process"):
        process_request_data: ProcessRequestData = ProcessRequestData("testing", "default", 1, "abc", "abc",
                                                                      {"abc": "test"})
        process_data: ProcessData = client.run_process_async(process_request_data)
        print(str(process_data.get_process_id()) + "," + str(process_data.get_status()))

    if os.getenv("rerun"):
        process_data : ProcessData = client.rerun(8150036)
        print(str(process_data))

    if os.getenv("search_process_executions"):
        process_search_request: ProcessSearchRequest = ProcessSearchRequest()
        process_search_request.set_process_ids([8200028, 8200025, 8200031])
        processes_data: list['ProcessData'] = client.search_process_executions(process_search_request)
        print(str(processes_data))

    if os.getenv("invoke_api_mapping_get"):
        response = client.invoke_api_mapping_get(endpoint="test-endpoint/gYSrAHFzoTf7areCaznA/MRZ7lPU8keONrvR7Sc88", api_call_type=ApiCallType.ASYNC)
        print(str(response))

    if os.getenv("invoke_api_mapping_post"):
        response = client.invoke_api_mapping_post(endpoint="test-endpoint/gYSrAHFzoTf7areCaznA/MRZ7lPU8keONrvR7Sc88", input={"testing": 123}, api_call_type=ApiCallType.ASYNC)
        print(str(response))

    if os.getenv("run_process"):
        process_data: ProcessData = client.run_process_sync(process_request_data)
        print(str(process_data.get_process_id()) + "," + str(process_data.get_status()))

    client_config.add_worker_profile(WorkerProfile("default", "class_worker", "class_worker", 5, False))

    ## Both parameters are optional if both are supplied then scanning will be happening considering both if one of them is applied scanning will happen considering that else nothing will be registered
    ## we will ensure there are no duplicate workers registered (by name)
    ## decorator will be scanned by current_directory_full_path
    ## for callables references registration will be happened by decorator if provided else by function name itself
    client.register_workers(current_directory_full_path)

    worker : Worker = Worker(manually_registered_worker, "manually_registered_worker")
    client.register_worker(worker)

    worker : Worker = Worker(test_multiple_outputs, "test_multiple_outputs")
    client.register_worker(worker)

    worker: Worker = Worker(task_second_worker, "task_second_worker")
    client.register_worker(worker)

    client.start()

def test_multiple_outputs(inp : dict):
    return 1, ["test1", "test2"], {"name", "test3"}, {"dictionary": "output1"}

def manually_registered_worker(inp : dict):
    return {
        "name": "Manually registered worker !!!"
    }

if __name__ == "__main__":
    main()