import json
import logging

import requests

from ..http.http_request_factory import HttpRequestFactory
from ..http.http_client_factory import HttpClientFactory
from ...common.step_size import StepSize
from ...common.work_request import WorkRequest
from ...configs.client_config import ClientConfig

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class PollerClient:
    CLIENTS_POLL_URL = "api/clients/poll"

    def __init__(self, client_config: ClientConfig, http_client_factory: HttpClientFactory,
                 http_request_factory: HttpRequestFactory):
        self.client_config = client_config
        self.http_client = http_client_factory.create()
        self.http_request_factory = http_request_factory

    def poll(self, step_sizes: list['StepSize']) -> list['WorkRequest']:
        if not step_sizes:
            step_sizes = []


        json_body = []
        for step_size in step_sizes:
            json_body.append({
                "stepQueueNameData": {
                    "orgId" : step_size.get_step_queue_name_data().get_org_id(),
                    "namespace": step_size.get_step_queue_name_data().get_namespace(),
                    "name": step_size.get_step_queue_name_data().get_name(),
                    "stepType": step_size.get_step_queue_name_data().get_step_type().name
                },
                "size": step_size.get_size()
            })
        params = {
            "size": str(self.client_config.get_work_request_batch_size())
        }

        try:
            response = self.http_request_factory.create_post_request(self.CLIENTS_POLL_URL, params=params,
                                                                     body=json_body)
            response.raise_for_status()  # This will raise an HTTPError for bad responses
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while polling: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred while polling: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred while polling: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e

        if not response.text.strip():
            logger.debug("Did not receive any work, continuing to poll.")
            return []

        try:
            work_requests_dict_list = json.loads(response.text)

            return [
                WorkRequest(
                    process_id=work_request_json.get("processId", None),
                    step_id=work_request_json["stepId"],
                    step_execution_id=work_request_json["stepExecutionId"],
                    step_name=work_request_json.get("stepName", None),
                    step_ref=work_request_json.get("stepRef", None),
                    step_namespace=work_request_json.get("stepNamespace", None),
                    input_param=work_request_json["inputParam"],
                    is_optional=work_request_json.get("isOptional", None),
                    polled=work_request_json["polled"],
                    scheduled=work_request_json["scheduled"],
                    updated=work_request_json["updated"],
                    priority=work_request_json["priority"],
                )
                for work_request_json in work_requests_dict_list
            ]
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err