import json
import logging
from typing import Any

import requests

from ..http.http_client_factory import HttpClientFactory
from ..http.http_request_factory import HttpRequestFactory
from ...common.api_call_type import ApiCallType
from ...common.process_action_response_data import ProcessActionResponseData
from ...common.process_data import ProcessData
from ...common.process_request_data import ProcessRequestData
from ...common.process_search_request import ProcessSearchRequest
from ...common.step_data import StepData
from ...common.unmeshed_constants import UnmeshedConstants
from...configs.client_config import ClientConfig

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ProcessClient:
    def __init__(self, http_client_factory: HttpClientFactory,http_request_factory: HttpRequestFactory, client_config: ClientConfig):
        self.client_config = client_config
        self.__http_client = http_client_factory.create()
        self.http_request_factory = http_request_factory
        self.http_request_factory = http_request_factory
        self.__run_process_request_url = "api/process/"

    def __convert_to_enum(self, enum_class, value):
        if isinstance(value, str) and hasattr(enum_class, value):
            return getattr(enum_class, value)
        return None

    def __populate_process_data(self, process_data_json) -> ProcessData:
        return ProcessData(
            process_id=process_data_json.get("processId"),
            process_type=self.__convert_to_enum(UnmeshedConstants.ProcessType, process_data_json.get("processType")),
            trigger_type=self.__convert_to_enum(UnmeshedConstants.ProcessTriggerType, process_data_json.get("triggerType")),
            namespace=process_data_json.get("namespace"),
            name=process_data_json.get("name"),
            status=self.__convert_to_enum(UnmeshedConstants.ProcessStatus, process_data_json.get("status")),
            version=process_data_json.get("version"),
            history_id=process_data_json.get("historyId"),
            request_id=process_data_json.get("requestId"),
            correlation_id=process_data_json.get("correlationId"),
            input=process_data_json.get("input"),
            output=process_data_json.get("output"),
            state=process_data_json.get("state"),
            secret_state=process_data_json.get("secretState"),
            auth_claims=process_data_json.get("authClaims"),
            step_id_count=process_data_json.get("stepIdCount"),
            shard_name=process_data_json.get("shardName"),
            shard_instance_id=process_data_json.get("shardInstanceId"),
            steps=process_data_json.get("steps"),
            step_records=process_data_json.get("stepRecords"),
            created=process_data_json.get("created"),
            updated=process_data_json.get("updated"),
            created_by=process_data_json.get("createdBy")
        )

    def __populate_processes_data(self, processes_data_json) -> list['ProcessData']:
        return [
            ProcessData(
                process_id=process_data.get("processId"),
                process_type=self.__convert_to_enum(UnmeshedConstants.ProcessType, process_data.get("processType")),
                trigger_type=self.__convert_to_enum(UnmeshedConstants.ProcessTriggerType,
                                                    process_data.get("triggerType")),
                namespace=process_data.get("namespace"),
                name=process_data.get("name"),
                status=self.__convert_to_enum(UnmeshedConstants.ProcessStatus, process_data.get("status")),
                version=process_data.get("version"),
                history_id=process_data.get("historyId"),
                request_id=process_data.get("requestId"),
                correlation_id=process_data.get("correlationId"),
                input=process_data.get("input"),
                output=process_data.get("output"),
                state=process_data.get("state"),
                secret_state=process_data.get("secretState"),
                auth_claims=process_data.get("authClaims"),
                step_id_count=process_data.get("stepIdCount"),
                shard_name=process_data.get("shardName"),
                shard_instance_id=process_data.get("shardInstanceId"),
                steps=process_data.get("steps"),
                step_records=process_data.get("stepRecords"),
                created=process_data.get("created"),
                updated=process_data.get("updated"),
                created_by=process_data.get("createdBy")
            ) for process_data in processes_data_json
        ]

    def __populate_step_data(self, step_data_json) -> StepData:
        return StepData(
            id=step_data_json.get("id"),
            process_id=step_data_json.get("processId"),
            ref=step_data_json.get("ref"),
            parent_id=step_data_json.get("parentId"),
            parent_ref=step_data_json.get("parentRef"),
            namespace=step_data_json.get("namespace"),
            name=step_data_json.get("name"),
            type=self.__convert_to_enum(UnmeshedConstants.StepType,
                                                step_data_json.get("type")),
            step_definition_history_id = step_data_json.get("stepDefinitionHistoryId"),
            status=self.__convert_to_enum(UnmeshedConstants.StepStatus,
                                                step_data_json.get("status")),
            input=step_data_json.get("input"),
            output=step_data_json.get("output"),
            worker_id=step_data_json.get("workerId"),
            start=step_data_json.get("start"),
            schedule=step_data_json.get("schedule"),
            priority=step_data_json.get("priority"),
            updated=step_data_json.get("updated"),
            optional=step_data_json.get("optional"),
            execution_list=step_data_json.get("executionList")
        )

    def __populate_process_action_response_data(self, bulk_termination_response) -> ProcessActionResponseData:
        return ProcessActionResponseData(
            count = bulk_termination_response.get("count"),
            details = bulk_termination_response.get("details")
        )

    def run_process_async(self, process_request_data: ProcessRequestData) -> ProcessData :
        params = {
            "clientId": self.client_config.get_client_id()
        }
        json_body = {
            "name" : process_request_data.get_name(),
            "namespace": process_request_data.get_namespace(),
            "version" : process_request_data.get_version(),
            "requestId": process_request_data.get_request_id(),
            "correlationId": process_request_data.get_correlation_id(),
            "input": process_request_data.get_input()
        }
        try:
            response = self.http_request_factory.create_post_request(self.__run_process_request_url + "runAsync",
                                                                     params=params,
                                                                     body=json_body)
            if response.status_code != 200:
                raise RuntimeError("Invalid process run request " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while running process: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred while running process: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred while running process: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            process_data_json = json.loads(response.text)
            return self.__populate_process_data(process_data_json)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err

    def run_process_sync(self, process_request_data: ProcessRequestData) -> ProcessData :
        params = {
            "clientId": self.client_config.get_client_id()
        }
        json_body = {
            "name": process_request_data.get_name(),
            "namespace": process_request_data.get_namespace(),
            "version": process_request_data.get_version(),
            "requestId": process_request_data.get_request_id(),
            "correlationId": process_request_data.get_correlation_id(),
            "input": process_request_data.get_input()
        }
        try:
            response = self.http_request_factory.create_post_request(self.__run_process_request_url + "runSync",
                                                                     params=params,
                                                                     body=json_body)
            if response.status_code != 200:
                raise RuntimeError("Invalid process run request " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while running process: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred while running process: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred while running process: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            process_data_json = json.loads(response.text)
            return self.__populate_process_data(process_data_json)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err

    def get_process_data(self, process_id : int, include_steps: bool) -> ProcessData:
        if process_id  is None:
            raise ValueError("Process ID cannot be None")

        url = self.__run_process_request_url + "context/" + str(process_id)
        params = {
            "includeSteps": include_steps
        }
        try:
            response = self.http_request_factory.create_get_request(url, params=params)
            if response.status_code != 200:
                raise RuntimeError("Invalid fetch process data request " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while fetching process data: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred fetching process data: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred fetching process data: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            process_data_json = json.loads(response.text)
            return self.__populate_process_data(process_data_json)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err

    def get_step_data(self, step_id : int) -> StepData:
        if step_id  is None:
            raise ValueError("Step ID cannot be None")
        url = self.__run_process_request_url + "stepContext/" + str(step_id)
        try:
            response = self.http_request_factory.create_get_request(url, params={})
            if response.status_code != 200:
                raise RuntimeError("Invalid fetch step data request " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while fetching step data: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred while fetching step data: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred while fetching step data: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            step_data_json = json.loads(response.text)
            return self.__populate_step_data(step_data_json)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err

    def bulk_terminate(self, process_ids: list['int'], reason: str) -> ProcessActionResponseData:
        if process_ids is None:
            raise ValueError("ProcessIds's cannot be None")
        url = self.__run_process_request_url + "bulkTerminate"
        params = {}
        if reason is not None:
            params = {
                "reason": reason
            }
        try:
            response = self.http_request_factory.create_post_request(url, params, process_ids)
            if response.status_code != 200:
                raise RuntimeError("Failed to bulk terminate " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while bulk termination: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred while bulk termination: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred while bulk termination: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            bulk_termination_response = json.loads(response.text)
            return self.__populate_process_action_response_data(bulk_termination_response)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err

    def bulk_resume(self, process_ids: list['int']) -> ProcessActionResponseData:
        if process_ids is None:
            raise ValueError("ProcessIds's cannot be None")
        url = self.__run_process_request_url + "bulkResume"
        params = {
            "clientId": self.client_config.get_client_id()
        }
        try:
            response = self.http_request_factory.create_post_request(url, params, process_ids)
            if response.status_code != 200:
                raise RuntimeError("Failed to bulk resume " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while bulk resume: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred while bulk resume: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred while bulk resume: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            bulk_resume_response = json.loads(response.text)
            return self.__populate_process_action_response_data(bulk_resume_response)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err


    def bulk_reviewed(self, process_ids: list['int'], reason: str) -> ProcessActionResponseData:
        if process_ids is None:
            raise ValueError("ProcessIds's cannot be None")
        url = self.__run_process_request_url + "bulkReviewed"
        params = {
            "clientId": self.client_config.get_client_id()
        }
        if reason is not None:
            params["reason"] = reason
        try:
            response = self.http_request_factory.create_post_request(url, params, process_ids)
            if response.status_code != 200:
                raise RuntimeError("Failed to bulk reviewed " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while bulk reviewed: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred while bulk reviewed: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred while bulk reviewed: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            bulk_reviewed_response = json.loads(response.text)
            return self.__populate_process_action_response_data(bulk_reviewed_response)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err

    def rerun(self, process_id : int, version: int) -> ProcessData:
        if process_id is None:
            raise ValueError("Process ID cannot be None")
        params = {
            "clientId": self.client_config.get_client_id(),
            "processId": process_id
        }
        url = self.__run_process_request_url + "rerun"
        if version is not None:
            params["version"]= version

        try:
            response = self.http_request_factory.create_post_request(url, params, None)
            if response.status_code != 200:
                raise RuntimeError("Failed to rerun request " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred rerunning process: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred while rerunning process: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred while rerunning process: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            process_data = json.loads(response.text)
            return self.__populate_process_data(process_data)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err


    def search_process_executions(self, params : ProcessSearchRequest) -> list['ProcessData']:
        query_params = {
            "startTimeEpoch": params.start_time_epoch if params.start_time_epoch else None,
            "endTimeEpoch": params.end_time_epoch if params.end_time_epoch else None,
            "namespace": params.namespace,
            "processTypes": ",".join(params.process_types) if params.process_types else None,
            "triggerTypes": ",".join(params.trigger_types) if params.trigger_types else None,
            "names": ",".join(params.names) if params.names else None,
            "processIds": ",".join(map(str, params.process_ids)) if params.process_ids else None,
            "correlationIds": ",".join(params.correlation_ids) if params.correlation_ids else None,
            "requestIds": ",".join(params.request_ids) if params.request_ids else None,
            "statuses": ",".join(params.statuses) if params.statuses else None,
            "limit": params.limit if params.limit != 0 else None,
            "offset": params.offset if params.offset != 0 else None,
        }

        query_params = {k: v for k, v in query_params.items() if v is not None}
        url = "api/stats/process/search"
        try:
            response = self.http_request_factory.create_get_request(url, params=query_params)
            if response.status_code != 200:
                raise RuntimeError("Invalid fetch processes data " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while fetching processes data: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred fetching processes data: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred fetching processes data: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            proceses_data_json = json.loads(response.text)
            return self.__populate_processes_data(proceses_data_json)
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err


    def invoke_api_mapping_get(self, endpoint: str, id : str, correlation_id : str, api_call_type: ApiCallType) -> dict['str', 'Any']:
        if endpoint is None:
            raise ValueError("Endpoint cannot be None")
        query_params = {
            "id": id if id else None,
            "correlationId": correlation_id if correlation_id else None,
            "apiCallType" : api_call_type.name if api_call_type else ApiCallType.ASYNC.name
        }

        query_params = {k: v for k, v in query_params.items() if v is not None}
        url = "api/call/" + endpoint
        try:
            response = self.http_request_factory.create_get_request(url, params=query_params)
            if response.status_code != 200:
                raise RuntimeError("Failed invoking webhook get request " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while invoking webhook get request: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred invoking webhook get request: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred invoking webhook get request: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            response = json.loads(response.text)
            return response

        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err

    def invoke_api_mapping_post(self, endpoint: str, input: dict['str', 'Any'], id : str, correlation_id : str, api_call_type: ApiCallType) -> dict['str', 'Any']:
        if endpoint is None:
            raise ValueError("Endpoint cannot be None")
        query_params = {
            "id": id if id else None,
            "correlationId": correlation_id if correlation_id else None,
            "apiCallType": api_call_type.name if api_call_type else ApiCallType.ASYNC.name
        }

        query_params = {k: v for k, v in query_params.items() if v is not None}
        url = "api/call/" + endpoint
        try:
            response = self.http_request_factory.create_post_request(url, query_params, input)
            if response.status_code != 200:
                raise RuntimeError("Failed invoking webhook post request " + response.text)
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            logger.error("HTTP error occurred while invoking webhook post request: %s", http_err)
            raise RuntimeError(f"HTTP error: {http_err}") from http_err
        except requests.exceptions.RequestException as req_err:
            logger.error("Request error occurred invoking webhook post request: %s", req_err)
            raise RuntimeError(f"Request error: {req_err}") from req_err
        except Exception as e:
            logger.error("An unexpected error occurred invoking webhook post request: %s", e)
            raise RuntimeError(f"Unexpected error: {e}") from e
        try:
            response = json.loads(response.text)
            return response
        except json.JSONDecodeError as json_err:
            logger.error("Failed to parse the response JSON: %s", json_err)
            raise RuntimeError(f"Failed to parse response JSON: {json_err}") from json_err