import logging
import time

from ..http.http_request_factory import HttpRequestFactory
from ..http.http_client_factory import HttpClientFactory
from ...common.worker_step_name import WorkerStepName
from ...configs.client_config import ClientConfig

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class RegistrationClient:
    CLIENTS_REGISTER_URL = "api/clients/register"

    def __init__(self, client_config: ClientConfig, http_client_factory: HttpClientFactory,
                 http_request_factory: HttpRequestFactory):
        self.client_config = client_config
        self.http_client = http_client_factory.create()
        self.request_factory = http_request_factory
        self.__worker_step_names: list['WorkerStepName'] = []

    def renew_registration(self) -> str:
        supported_steps = []
        for worker_step_name in self.__worker_step_names:
            step_queue_name_data = worker_step_name.get_step_queue_name_data()
            worker_profile = worker_step_name.get_worker_profile()

            worker_data = {
                "orgId": step_queue_name_data.get_org_id(),
                "namespace": worker_profile.get_namespace(),
                "stepType": step_queue_name_data.get_step_type().name,
                "name": worker_profile.get_worker_queue_name()
            }
            supported_steps.append(worker_data)

        logger.info("Renewing registration for the following workers: %s", supported_steps)

        params = {}
        retries = 5
        delay = 2

        for attempt in range(retries):
            try:
                response = self.request_factory.create_put_request(
                    self.CLIENTS_REGISTER_URL, params=params, body=supported_steps
                )
                logger.debug("Response from server: %s", response)

                if response.status_code != 200:
                    logger.error("Did not receive 200! %s", response.text)
                    raise Exception(
                        f"Got a non-200 result from the server when trying to renew registration: {response.status_code}"
                    )

                logger.info("Successfully renewed registration for workers.")
                return response.text

            except Exception as e:
                logger.error(
                    "An error occurred while renewing registration (Attempt %d/%d): %s",
                    attempt + 1, retries, str(e)
                )

            if attempt < retries - 1:
                logger.info("Retrying in %d seconds...", delay)
                time.sleep(delay)

        raise RuntimeError("Registration failed after multiple attempts.")

    def set_worker_step_names(self, worker_step_names: list['WorkerStepName']):
        self.__worker_step_names = worker_step_names

    def get_worker_step_names(self) -> list['WorkerStepName']:
        return self.__worker_step_names