import atexit
import logging
import json
import time
import threading
from concurrent.futures import ThreadPoolExecutor
from queue import Queue

from ..http.http_request_factory import HttpRequestFactory
from ..http.http_client_factory import HttpClientFactory
from ...common.client_submit_result import ClientSubmitResult
from ...common.unmeshed_resource_pool import UnmeshedResourcePool
from ...common.work_response import WorkResponse
from ...common.work_response_tracker import WorkResponseTracker
from ...configs.client_config import ClientConfig

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

class SubmitClient:
    CLIENTS_RESULTS_URL = "api/clients/bulkResults"

    def __init__(self, http_client_factory: HttpClientFactory, http_request_factory: HttpRequestFactory, client_config: ClientConfig):
        self.http_client = http_client_factory.create()
        self.http_request_factory = http_request_factory
        self.client_config = client_config

        if not self.client_config.get_client_id():
            raise ValueError("Cannot submit results without a clientId")

        self.__main_queue = Queue(maxsize=100000)
        self.__retry_queue = Queue(maxsize=100000)
        self.__submit_tracker : dict[int, WorkResponseTracker] = {}
        self.__submit_tracker_lock = threading.Lock()

        self.__executor = ThreadPoolExecutor(max_workers=3)
        atexit.register(self.__executor.shutdown, wait=False)
        self.__executor.submit(self.__cleanup_lingering_submit_trackers)
        self.__executor.submit(self.__process_queue, self.__main_queue, False)
        self.__executor.submit(self.__process_queue, self.__retry_queue, True)


    def __cleanup_lingering_submit_trackers(self):
        while True:
            counter = 0
            current_millis = int(time.time() * 1000)

            with self.__submit_tracker_lock:
                step_ids = list(self.__submit_tracker.keys())
                for step_id in step_ids:
                    tracker = self.__submit_tracker.get(step_id)
                    if isinstance(tracker, WorkResponseTracker):
                        seconds = 10 * 60
                        remove_eligible = current_millis - tracker.get_queued_time() > seconds * 1000
                        if remove_eligible:
                            tracker.get_resource_pool().release()
                            counter += 1
                            del self.__submit_tracker[step_id]

            time.sleep(5)

    def __process_queue(self, queue, is_retry):
        while True:
            batch = []
            for _ in range(self.client_config.get_response_submit_batch_size()):
                if not queue.empty():
                    batch.append(queue.get())

            if not batch:
                time.sleep(1)
                continue

            logging.debug(f"Processing batch of size {len(batch)} from {'retryQueue' if is_retry else 'mainQueue'}")

            params = {}
            json_body = []
            for response in batch:
                json_body.append({
                    "processId": response.get_process_id(),
                    "stepId": response.get_step_id(),
                    "startedAt": response.get_started_at(),
                    "stepExecutionId": response.get_step_execution_id(),
                    "output": response.get_output(),
                    "status": response.get_status(),
                    "rescheduleAfterSeconds": response.get_reschedule_after_seconds()
                })
            try:
                response = self.http_request_factory.create_post_request(self.CLIENTS_RESULTS_URL, params=params, body=json_body)
            except Exception as e:
                time.sleep(3)
                logging.error(f"Bulk request failed for batch. Re-queuing all items. Error: {str(e)}")
                for work_response in batch:
                    self.__handle_all_request_failure(work_response, str(e))
                continue

            logging.debug(f"Bulk request response: {response.status_code}")

            if response.status_code != 200:
                logging.error(f"Bulk request failed with status {response.status_code}. Re-queuing all items.")
                for work_response in batch:
                    self.__handle_all_request_failure(work_response, "Response status not 200")
                continue

            response_map = json.loads(response.text)

            for work_response in batch:
                result = response_map.get(str(work_response.get_step_id()))
                with self.__submit_tracker_lock:
                    work_response_tracker = self.__submit_tracker.get(work_response.get_step_id())

                if not result or result.get("error_message"):
                    logging.error(
                        f"Error for WorkResponse {work_response.get_process_id()} {work_response.get_step_id()}: {result.get('error_message', 'No result')}")
                    self.__enqueue_for_retry(work_response, result, work_response_tracker)
                else:
                    logging.info(f"Result from stepId {work_response.get_step_id()} submitted!")
                    with self.__submit_tracker_lock:
                        self.__submit_tracker.pop(work_response.get_step_id(), None)
                    if work_response_tracker:
                        work_response_tracker.get_resource_pool().release()


    def __handle_all_request_failure(self, work_response: WorkResponse, message):
        with self.__submit_tracker_lock:
            work_response_tracker = self.__submit_tracker.get(work_response.get_step_id())

        if work_response_tracker:
            self.__enqueue_for_retry(work_response, ClientSubmitResult(work_response.get_process_id(), work_response.get_step_id(), message, 400), work_response_tracker)

    def __enqueue_for_retry(self, work_response: WorkResponse, result, work_response_tracker: WorkResponseTracker):
        if self.__is_permanent_error__(result.get_error_message()):
            logging.error(f"Permanent error for WorkResponse {work_response.get_process_id()}: {result.get_error_message()}")
            with self.__submit_tracker_lock:
                self.__submit_tracker.pop(work_response.get_step_id(), None)
            work_response_tracker.get_resource_pool().release()
            return

        count = work_response_tracker.get_retry_count() + 1
        if count > self.client_config.get_max_submit_attempts():
            logging.error(f"Max retry attempts reached for WorkResponse {work_response.get_step_id()} - {work_response.get_process_id()}")
            with self.__submit_tracker_lock:
                self.__submit_tracker.pop(work_response.get_step_id(), None)
            work_response_tracker.get_resource_pool().release()
            return

        work_response_tracker.set_retry_count(count)
        self.__retry_queue.put(work_response)
        logging.debug(f"Re-queued WorkResponse {work_response.get_process_id()} for retry attempt {count}")

    def __is_permanent_error__(self, error_message):
        if not error_message:
            return False
        for keyword in self.client_config.permanent_error_keywords():
            if keyword in error_message:
                return True
        return False

    def submit(self, work_response: WorkResponse, resource_pool: UnmeshedResourcePool):
        logger.debug(f"Submitting results to queue: {work_response}")
        with self.__submit_tracker_lock:
            self.__submit_tracker[work_response.get_step_id()] = WorkResponseTracker(work_response, 0, int(time.time() * 1000), resource_pool)
            self.__main_queue.put(work_response)
            logger.info(f"Result[{work_response.get_status()}] from stepId {work_response.get_step_id()} queued!")

    def get_submit_tracker_size(self) -> int:
        with self.__submit_tracker_lock:
            return len(self.__submit_tracker)
