import threading
import time


### When resource release decrease count for that process else on acquire increment count
class UnmeshedResourcePool:

    def __init__(self, permits: int, max_permits: int):
        # Adding a max limit for available permits
        self._resources: dict['str', 'int'] = {"available_permits": permits}
        self._max_permits = max_permits  # Maximum limit of permits
        self._lock = threading.Lock()

    def try_acquire(self, permits: int, timeout: float, unit: str) -> bool:
        """
        Attempts to acquire the specified number of permits within the given timeout.
        Returns True if permits are acquired, False if not within the timeout.
        """
        # Convert timeout to seconds if necessary
        if unit == 'milliseconds':
            timeout = timeout / 1000.0  # Convert to seconds
        elif unit == 'seconds':
            pass
        else:
            raise ValueError(f"Unsupported time unit: {unit}")

        start_time = time.time()
        with self._lock:
            while time.time() - start_time < timeout:
                if self._resources["available_permits"] >= permits:
                    self._resources["available_permits"] -= permits
                    return True
                time.sleep(0.01)  # Sleep a bit before checking again
        return False

    def acquire(self, timeout=None) -> bool:
        with self._lock:
            if self._resources["available_permits"] > 0:
                self._resources["available_permits"] -= 1
                return True
            else:
                start_time = time.time()
                while time.time() - start_time < timeout:
                    if self._resources["available_permits"] > 0:
                        self._resources["available_permits"] -= 1
                        return True
                return False

    def release(self, n: int = 1):
        with self._lock:
            # Only allow releasing if the current available permits won't exceed the max
            if int(self._resources["available_permits"]) + n <= self._max_permits:
                self._resources["available_permits"] = int(self._resources["available_permits"]) + n
            else:
                # Ensure the available permits do not exceed the maximum limit
                self._resources["available_permits"] = self._max_permits

    def available_permits(self) -> int:
        with self._lock:
            return self._resources["available_permits"]