import atexit
import logging
import math
import random
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from typing import get_origin, Type, Callable
import inspect

from .apis.http.http_client_factory import HttpClientFactory
from .apis.http.http_request_factory import HttpRequestFactory
from .apis.poller.poller_client import PollerClient
from .apis.process.process_client import ProcessClient
from .apis.registration.registration_client import RegistrationClient
from .apis.submit.submit_client import SubmitClient
from .apis.submit.work_response_builder import WorkResponseBuilder
from .apis.workers.worker import Worker
from .common.api_call_type import ApiCallType
from .common.process_action_response_data import ProcessActionResponseData
from .common.process_data import ProcessData
from .common.process_request_data import ProcessRequestData
from .common.process_search_request import ProcessSearchRequest
from .common.step_data import StepData
from .common.step_queue_name_data import StepQueueNameData
from .common.step_queue_poll_state import StepQueuePollState
from .common.step_size import StepSize
from .common.unmeshed_constants import UnmeshedConstants
from .common.unmeshed_resource_pool import UnmeshedResourcePool
from .common.work_request import WorkRequest
from .common.work_response import WorkResponse
from .common.worker_profile import WorkerProfile
from .common.worker_step_name import WorkerStepName
from .configs.client_config import ClientConfig
from .schedulers.step_result import StepResult
from .utils.worker_scanner import WorkerScanner

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class UnmeshedClient:
    __thread_local = threading.local()

    def __init__(self, client_config : ClientConfig):
        self.workers = list()
        self.client_config = client_config
        self.__poll_states: dict[str, 'StepQueuePollState'] = dict()
        self.__permits_zero = 0
        self.__polling_lock = threading.Lock()
        self.__run_and_submit_step_lock = threading.Lock()
        self.__poll_size_zero_time = 0
        self.__executing_count = 0
        self.__last_printed = 0
        self.__polling_error_reported = False
        self.__jitter_upper_bound_in_millis = 2000
        self.__initial_delay_millis = 50

        self.__backoff_multiplier = 2
        self.__retry_count = 0

        self.__http_client_factory = HttpClientFactory(client_config)
        self.__http_request_factory = HttpRequestFactory(client_config)

        self.__registration_client = RegistrationClient(client_config, self.__http_client_factory,  self.__http_request_factory)
        self.__poller_client = PollerClient(client_config, self.__http_client_factory,  self.__http_request_factory)
        self.__submit_client = SubmitClient(self.__http_client_factory,  self.__http_request_factory, self.client_config)
        self.__process_client = ProcessClient(self.__http_client_factory,  self.__http_request_factory, client_config)
        self.__work_response_builder = WorkResponseBuilder()
        self.__executor = ThreadPoolExecutor(max_workers=min(2, client_config.get_max_threads_count()))
        atexit.register(self.__executor.shutdown, wait=False)

    def __schedule_polling(self):
        try:
            while True:
                self.__poll_and_submit_work_results()
                time.sleep(self.__delay_in_millis() / 1000)
        except Exception as ex:
            self.__retry_count += 1
            time.sleep(self.__delay_in_millis() / 1000)

    @staticmethod
    def __formatted_worker_id(namespace: str, name: str) -> str:
        return f"{namespace}:-#-:{name}"

    def run_process_sync(self, process_request_data: ProcessRequestData) -> ProcessData:
        return self.__process_client.run_process_sync(process_request_data)

    def run_process_async(self, process_request_data: ProcessRequestData) -> ProcessData:
        return self.__process_client.run_process_async(process_request_data)

    def get_process_data(self, process_id: int, include_steps: bool = False) -> ProcessData:
        return self.__process_client.get_process_data(process_id, include_steps)

    def get_step_data(self, step_id: int) -> StepData:
        return self.__process_client.get_step_data(step_id)

    def bulk_terminate(self,  process_ids: list['int'], reason: str = None) -> ProcessActionResponseData:
        return self.__process_client.bulk_terminate(process_ids, reason)

    def bulk_resume(self, process_ids: list['int']) -> ProcessActionResponseData:
        return self.__process_client.bulk_resume(process_ids)

    def bulk_reviewed(self,  process_ids: list['int'], reason: str = None) -> ProcessActionResponseData:
        return self.__process_client.bulk_reviewed(process_ids, reason)

    def rerun(self, process_id: int, version: int = None) -> ProcessData:
        return self.__process_client.rerun(process_id, version)

    def search_process_executions(self, params: ProcessSearchRequest) -> list['ProcessData']:
        return self.__process_client.search_process_executions(params)

    def invoke_api_mapping_get(self, endpoint: str = None, id: str = None, correlation_id: str = None, api_call_type: ApiCallType = None) -> dict['str', 'Any']:
        return self.__process_client.invoke_api_mapping_get(endpoint, id, correlation_id, api_call_type)

    def invoke_api_mapping_post(self, endpoint: str = None, input: dict['str', 'Any'] = None, id: str = None, correlation_id: str = None, api_call_type: ApiCallType = None) -> dict['str', 'Any']:
        return self.__process_client.invoke_api_mapping_post(endpoint, input, id, correlation_id, api_call_type)

    def register_worker(self, workerToScan : Worker):
        workersToScan: list = [workerToScan]
        self.__scan_and_register_workers(workersToScan)

    def register_workers(self, scan_path: str):
        workersToScan : list = WorkerScanner.find_workers(scan_path);
        self.__scan_and_register_workers(workersToScan)

    def __scan_and_register_workers(self, scanned_workers : list[Worker]):
        worker_profiles = self.client_config.get_worker_profiles() or []
        self.workers.extend(scanned_workers)
        step_names = []

        for worker in self.workers:
            method = worker.get_execution_method()
            sig = inspect.signature(method)
            params = sig.parameters
            if len(params) != 1:
                continue
            matching_profiles = [profile for profile in worker_profiles if profile.get_worker_name() == worker.get_name()]
            if matching_profiles:
                for profile in matching_profiles:
                    step_names.append(WorkerStepName(
                        StepQueueNameData(1, profile.get_namespace(), UnmeshedConstants.StepType.WORKER,
                                          profile.get_worker_queue_name()),
                        profile
                    ))
            else:
                default_profile = WorkerProfile(worker.get_namespace(), worker.get_name(), worker.get_name(),
                                                worker.get_max_in_progress(), worker.is_io_thread())
                step_names.append(WorkerStepName(
                    StepQueueNameData(1, worker.get_namespace(), UnmeshedConstants.StepType.WORKER, worker.get_name()),
                    default_profile
                ))

        for step in step_names:
            logging.info(f"Registering worker {step}")

        self.__registration_client.set_worker_step_names(step_names)

    @staticmethod
    def __renew_registration_with_retry(renew_registration_task):
        retry_count = 0
        delay = 1
        max_delay = 10

        while True:
            try:
                logger.info(f"Attempting to renew registration. Retry count: {retry_count}")
                renew_registration_task()
                logger.info("Registration renewed successfully.")
                break
            except Exception as e:
                retry_count += 1
                logger.error(f"Retry {retry_count} failed: {type(e).__name__}: {str(e)}")

            try:
                logger.info(f"Waiting for {delay} seconds before retrying...")
                time.sleep(delay)
            except InterruptedError:
                logger.error("Retry interrupted")
                raise RuntimeError("Retry interrupted") from e

            delay = min(delay + 2, max_delay)

    def start(self):
        if not self.client_config.has_token():
            logger.warning("Credentials not configured correctly. Have you passed in UNMESHED_AUTH_ID and UNMESHED_AUTH_TOKEN env vars?")

        if self.__registration_client.get_worker_step_names() is None or len(self.__registration_client.get_worker_step_names()) == 0:
            logger.error("No workers configured in this package. Will not poll for any work.")
            return

        for s in self.__registration_client.get_worker_step_names():
            default_max_size = 1000
            if s.get_worker_profile() is not None:
                default_max_size = s.get_worker_profile().get_max_in_progress()

            step_name_data : StepQueueNameData = s.get_step_queue_name_data()
            worker_id = self.__formatted_worker_id(step_name_data.get_namespace(),
                                                   step_name_data.get_name())
            self.__poll_states[worker_id] = StepQueuePollState(s, UnmeshedResourcePool(default_max_size, default_max_size), default_max_size)

        renew_registration_task = self.__registration_client.renew_registration
        self.__renew_registration_with_retry(renew_registration_task)


        logger.info("Unmeshed Python SDK started")
        self.__executor.submit(self.__schedule_polling)

    def __poll_and_submit_work_results(self):
        with self.__polling_lock:
            self.__retry_count = 0
            step_sizes: list[StepSize] = list()
            actual_sizes: dict[str, int] = dict()
            try:
                try:
                    for sps in self.__poll_states.values():
                        max_available = min(5000,
                                            min(self.client_config.get_work_request_batch_size(),
                                                sps.get_permits().available_permits()))
                        acquired = sps.get_permits().try_acquire(max_available, 250, "milliseconds")
                        if acquired:
                            s: StepQueueNameData = sps.get_worker_step_name().get_step_queue_name_data()
                            actual_sizes[self.__formatted_worker_id(s.get_namespace(), s.get_name())] = max_available
                            step_sizes.append(StepSize(s, max_available))
                except:
                    step_sizes = []

                if any(p.get_size() > 0 for p in step_sizes):
                    self.__permits_zero = 0
                    current_millis = int(time.time() * 1000)
                    self.__poll_size_zero_time = current_millis
                else:
                    current_millis = int(time.time() * 1000)
                    if self.__poll_size_zero_time < (current_millis - 1000):
                        self.__poll_size_zero_time = current_millis
                        self.__permits_zero += 1
                        if self.__permits_zero > 90:
                            logger.info("No polling for 90 seconds, will clear out permits")
                            for sps in self.__poll_states.values():
                                sps.get_permits().release(
                                    sps.get_total_permits() - sps.get_permits().available_permits())
                            self.__permits_zero = 0

                work_requests: list[WorkRequest] = self.__poller_client.poll(step_sizes)
                logger.debug("Received work requests : %s", len(work_requests))

                if self.__polling_error_reported:
                    self.__polling_error_reported = False
                    logger.info("Polling for work resumed successfully");

                for work_request in work_requests:
                    step_id: str = self.__formatted_worker_id(work_request.get_step_namespace(),
                                                              work_request.get_step_name())
                    if step_id in actual_sizes:
                        actual_sizes[step_id] = actual_sizes[step_id] - 1
                    self.__executing_count = self.__executing_count + 1
                    worker = next((w for w in self.workers if w.get_name() == work_request.get_step_name()), None)
                    if worker:
                        self.__run_step(worker, work_request)
                    else:
                        logger.error("No worker found for step '%s'", work_request.get_step_name())
            except Exception as e:
                exception_string = str(e)
                if self.__polling_error_reported == False:
                    if "java.net.ConnectException" in exception_string:
                        logger.error("Unmeshed client polling error %s", exception_string)
                    else:
                        logger.error("Unmeshed client polling error %s", e)
                self.__polling_error_reported = True
                logger.error("Error caught during polling & executing", e);
                self.__retry_count += 1
            finally:
                for step_id, un_used_permits in actual_sizes.items():
                    self.__poll_states[step_id].get_permits().release(un_used_permits)

            if time.time() * 1000 - self.__last_printed > 1000:
                log = ", ".join([
                    f"{s.get_worker_step_name().get_step_queue_name_data().get_namespace()}:{s.get_worker_step_name().get_step_queue_name_data().get_name()} = Available[{s.get_permits().available_permits()}] / [{s.get_total_permits()}]"
                    for s in self.__poll_states.values()])

                logger.info(
                    f"Currently running: {self.__executing_count} st: {self.__submit_client.get_submit_tracker_size()} t: {self.__executing_count + self.__submit_client.get_submit_tracker_size()} - permits {log} jobs {actual_sizes}")

                self.__last_printed = time.time() * 1000

    @staticmethod
    def get_current_work_request() -> WorkRequest:
        work_request: WorkRequest = UnmeshedClient.__thread_local.current_work_request
        return work_request

    def __run_step(self, worker: Worker, work_request: WorkRequest) -> None:
        with self.__run_and_submit_step_lock:
            method = worker.get_execution_method()
            sig = inspect.signature(method)
            params = sig.parameters

            param_name, param = next(iter(params.items()))
            expected_type: Type = param.annotation

            input_param = work_request.get_input_param()
            try:
                input_param = self.__convert_input_param(input_param, expected_type)
                UnmeshedClient.__thread_local.current_work_request = work_request

                result = method(input_param)


                step_result = StepResult(result if result is not None else dict())

                self.__handle_work_completion(work_request, step_result, None)

            except Exception as e:
                logger.error("Error executing step for work request %s: %s", work_request, e)
                self.__handle_work_completion(work_request, {"result": "Failed"}, e)

    @staticmethod
    def __convert_input_param(input_param, expected_type: Type):
        if get_origin(expected_type) is not None:
            if isinstance(input_param, dict):
                input_param = expected_type(**input_param)
            if not isinstance(input_param, expected_type):
                raise TypeError(f"Expected parameter of type {expected_type}, got {type(input_param)}")

        # Handle basic types
        if not isinstance(input_param, expected_type):
            if not (expected_type is dict and isinstance(input_param, str)):
                try:
                    input_param = expected_type(input_param)
                except Exception as e:
                    pass

            else:
                try:
                    input_param = eval(input_param)
                except ValueError:
                   raise TypeError(f"Invalid dictionary string: {input_param}")

        return input_param

    def __handle_work_completion(self, work_request: WorkRequest, step_result: StepResult, throwable: Exception):
        wr : WorkResponse = self.__work_response_builder.fail_response(work_request, throwable) if throwable else self.__work_response_builder.success_response(work_request, step_result)
        step_id = self.__formatted_worker_id(work_request.get_step_namespace(), work_request.get_step_name())
        self.__submit_client.submit(wr, self.__poll_states[step_id].get_permits())
        self.__executing_count = self.__executing_count - 1

    def __delay_in_millis(self):
        backoff = min(30000, round(self.__initial_delay_millis * math.pow(self.__backoff_multiplier, self.__retry_count)))

        # If this is a retry, add some jitter.
        if self.__retry_count != 0:
            backoff += random.randint(0, self.__jitter_upper_bound_in_millis - 1)

        return backoff